/* Communication: Main Interface
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <string.h>
#include <os_arm_bare_registers_def.h>
#include <os_arm_bare_inithardware.h>
#include <os_arm_bare_cpufreq.h>
#include "communication_buffer.h"
#include "../debug.h"
#include "../common.h"
#include "../led.h"
#include "../task.h"
#include "../flash.h"
#include "../utils.h"
#include "../config.h"
#include <pmic.h>
#include "../pmic_interrupt.h"
#include "../image.h"
#include "../memexec.h"

/** \defgroup communication Communication Subsystem
 * @{
 */

/** \file communication.c
 * \brief Communication Subsystem
 */

static void eraseFlash();
static void programFlash();
static bool setAddress(uint32_t address, uint32_t length);

/** \brief Commands from Host */
enum CommandType {
	CMD_RESET = 0,
	CMD_ADDRESS = 1,
	CMD_ERASE = 2,
	CMD_UPLOAD = 3,
	CMD_UPLOAD_FINISH = 4,
	CMD_REBOOT = 5,
	CMD_EXECUTE_RAM = 6,
};

/** \brief States */
enum StateType {
	STATE_INIT,
	STATE_IDLE,
	STATE_ERASE,
	STATE_UPLOADING,
	STATE_ERROR,
};

#ifdef DEBUG_COMM_STATES
/** \brief state names */
static const char * stateNames[] = {
	[STATE_INIT]		= "INIT",
	[STATE_IDLE]		= "IDLE",
	[STATE_ERASE]		= "ERASE",
	[STATE_UPLOADING]	= "UPLOADING",
	[STATE_ERROR]		= "ERROR",
};
#endif

/** \brief true iff system is busy (packets are discared) */
static bool busy = FALSE;

/** \brief Current State */
static enum StateType state = STATE_INIT;

/** \brief Replied Status Message OK */
static const aligned32 struct PacketHeader STAT_OK =
	{ .length = COMM_PKT_HEADER_SIZE, .cmd = 0 };
/** \brief Replied Status Error "Address or length invalid" */
static const aligned32 struct PacketHeader STAT_ERR_ADDR =
	{ .length = COMM_PKT_HEADER_SIZE, .cmd = 1 };
/** \brief Replied Status Error "Block unlocking failed" */
static const aligned32 struct PacketHeader STAT_ERR_UNLOCK =
	{ .length = COMM_PKT_HEADER_SIZE, .cmd = 2 };
/** \brief Replied Status Error "Block erase failed" */
static const aligned32 struct PacketHeader STAT_ERR_ERASE =
	{ .length = COMM_PKT_HEADER_SIZE, .cmd = 3 };
/** \brief Replied Status Error "Block programming failed" */
static const aligned32 struct PacketHeader STAT_ERR_PROGRAM =
	{ .length = COMM_PKT_HEADER_SIZE, .cmd = 4 };
/** \brief Replied Status Error "Unknown error" */
static const aligned32 struct PacketHeader STAT_ERR_UNKNOWN =
	{ .length = COMM_PKT_HEADER_SIZE, .cmd = 5 };

/** \brief Address to be flashed */
static uint32_t flashAddress;
/** \brief Length of data to be flashed */
static uint32_t flashLength;
/** \brief current position in flash */
static uint32_t flashPos;

static bool imageRamExec = FALSE;
static void * imageRamAddress;
static uint32_t imageRamLength;

/** \brief Data to be flashed */
static const void * curData;
/** \brief current length of data to be flashed */
static uint32_t curLength;

/** \brief switch state of communication automaton
 * \param newState new state to switch to */
static inline void switchState(enum StateType newState) {
#ifdef DEBUG_COMM_STATES
	/* trace states if flag is defined */
	if (state != newState) {
		debugf("Switching from state %s to state %s", stateNames[state],
			stateNames[newState]);
	}
#endif
#if defined(POWER_UP_UPLOADING) && POWER_UP_UPLOADING == TRUE
	if (newState == STATE_UPLOADING && state != STATE_UPLOADING) {
		/* TODO: stabilze clocks */
		cpufreqSetFrequency(SEnF_ARM_CPUFREQ_416);
	} else if (state == STATE_UPLOADING && newState != STATE_UPLOADING) {
		cpufreqSetFrequency(SEnF_ARM_CPUFREQ_104);
	}
#endif
	busy = FALSE;
	state = newState;
}

/** \brief send error code and switch to error state
 * \param pkt error packet */
static inline void err(const struct PacketHeader * pkt) {
	switchState(STATE_ERROR);
	ledOn(LED_RED);
	writePMIC(PMIC_LED1_CONTROL, 0);
	communicationSendPacket(pkt);
}

/** \brief reset automaton and send OK Reply */
static inline void reset() {
	switchState(STATE_IDLE);
	/* turn LEDs off */
	writePMIC(PMIC_LED1_CONTROL, 0);
	ledOff(LED_RED);
	communicationSendPacket(&STAT_OK);
}

#ifdef DEBUG_COMM
/** \brief communication debug output
 * \param format format
 * \param args arguments
 * \note just a wrapper which is only active when DEBUG_COMM is defined
 * \see debugf */
#define commdebugf(format, args...) debugf(format, ##args)
#else
/** \brief communication debug output
 * \note just a dummy macro to ignore all debug statements if communication
 *  debug is disable */
#define commdebugf(format, args...)
#endif

/** \brief Process received packet
 * \param pkt packet */
void communicationProcessPacket(const struct PacketHeader * pkt,
	const void * payload, uint32_t payloadLen) {
	static bool sendLedState = FALSE;
	
	commdebugf("processing packet with length %d (%d payload)", pkt->length,
		payloadLen);
	
	/* toggle send led */
	if (sendLedState) {
		ledOff(LED_GREEN);
	} else {
		ledOn(LED_GREEN);
	}
	sendLedState = !sendLedState;
	
	if (busy) {
		commdebugf("Discarding packet with cmd %d, busy", pkt->cmd);
		return;
	}
	
	/* switch state */
	switch (state) {
		case STATE_INIT:
			/* init state */
			commdebugf("State: INIT, cmd: %d", pkt->cmd);
			/* stop boot timeout */
			utilStopBootTimeout();
			
			/* turn off red led (e.g. if image was invalid) */
			ledOff(LED_RED);
			
			switch (pkt->cmd) {
				case CMD_RESET:
					/* reset command: reset to state idle */
					commdebugf("\tcmd: RESET, reply: OK");
					reset();
					break;
				default:
					/* unknown command */
					commdebugf("\tcmd: ??, reply: errUnknown");
					err(&STAT_ERR_UNKNOWN);
					break;
			}
			break;
		case STATE_IDLE:
			/* idle state */
			commdebugf("State: IDLE, cmd: %d", pkt->cmd);
			switch (pkt->cmd) {
				case CMD_RESET:
					/* reset command: reset to state idle */
					commdebugf("\tcmd: RESET, reply: OK");
					reset();
					break;
				case CMD_EXECUTE_RAM:
					/* execute image from ram */
					commdebugf("\tcmd: EXECUTE_RAM, executing...");
					if (!imageRamExec) {
						commdebugf("\tERROR: no image loaded to ram");
						err(&STAT_ERR_UNKNOWN);
						break;
					} else {
						communicationSendPacket(&STAT_OK);
						/* do it */
						memExec(imageRamAddress, imageRamLength);
					}
				case CMD_REBOOT:
					/* reboot command: reboot platform */
					commdebugf("\tcmd: REBOOT, rebooting...");
					reboot();
					break;
				case CMD_ADDRESS:
					/* address command: get address and length of data */
					{
						const struct PacketAddress * addr =
							(const struct PacketAddress *)payload;
						commdebugf("\tcmd: ADDR, address: 0x%08x, length: %d",
							addr->address, addr->length);
						if (!setAddress(addr->address, addr->length)) {
							err(&STAT_ERR_ADDR);
						} else {
							communicationSendPacket(&STAT_OK);
							switchState(STATE_ERASE);
						}
					}
					break;
				default:
					/* unknown command */
					debugf("\tcmd: ??, reply: errUnknown");
					err(&STAT_ERR_UNKNOWN);
					break;
			}
			break;
		case STATE_ERASE:
			/* erase state */
			commdebugf("State: ERASE, cmd: %d", pkt->cmd);
			switch (pkt->cmd) {
				case CMD_RESET:
					/* reset command: reset to state idle */
					commdebugf("\tcmd: RESET, reply: OK");
					reset();
					break;
				case CMD_ERASE:
					/* erase command: erase given address */
					commdebugf("\tcmd: ERASE, reply deferred");
					/* discard further packets */
					busy = TRUE;
					/* enqueue erase task */
					taskAdd(&eraseFlash, FALSE);
					break;
				default:
					/* unknown command */
					commdebugf("\tcmd: ??, reply: errUnknown");
					err(&STAT_ERR_UNKNOWN);
					break;
			}
			break;
		case STATE_UPLOADING:
			/* uploading state */
			commdebugf("State: UPLOADING, cmd: %d", pkt->cmd);
			switch (pkt->cmd) {
				case CMD_RESET:
					/* reset command: reset to state idle */
					commdebugf("\tcmd: RESET, reply: OK");
					reset();
					break;
				case CMD_UPLOAD:
					/* upload command: upload data */
					commdebugf("\tcmd: UPLOAD, curLength = %d",
						payloadLen);
					curLength = payloadLen;
					curData = payload;
					/* discard further packets */
					busy = TRUE;
					/* enqueue program task */
					taskAdd(&programFlash, FALSE);
					break;
				case CMD_UPLOAD_FINISH:
					commdebugf("\tcmd: UPLOAD_FINISH");
					if (flashLength != flashPos) {
						commdebugf("\tProgram not fully received, reply: errUnknown");
						err(&STAT_ERR_UNKNOWN);
					} else {
						commdebugf("\tAll done, returning to idle, reply: OK");
						switchState(STATE_IDLE);
						communicationSendPacket(&STAT_OK);
					}
					break;
				default:
					commdebugf("\tcmd: ??, reply: errUnknown");
					err(&STAT_ERR_UNKNOWN);
					break;
			}
			break;
		case STATE_ERROR:
			commdebugf("State: ERROR, cmd: %d", pkt->cmd);
			switch (pkt->cmd) {
				case CMD_REBOOT:
					commdebugf("\tcmd: REBOOT, rebooting...");
					reboot();
					break;
				default:
					commdebugf("\tcmd: ??, reply: errUnknown");
					err(&STAT_ERR_UNKNOWN);
					break;
			}
			break;
		default:
			commdebugf("State: ??, reply: errUnknown");
			err(&STAT_ERR_UNKNOWN);
			break;
	}
}

/** \brief check given \a address and \a length
 *
 * - The given \a address is compared agains #COMM_FLASH_PROG_MIN which is a
 *    lower boundary for the program.
 * - The \a length + \a address is checked to not exceed the #FLASH_SIZE
 * - If they are valid, the ::flashAddress and ::flashLength fields are set
 * \param address address
 * \param length length
 * \return \c TRUE iff the address and length are valid */
bool setAddress(uint32_t address, uint32_t length) {
	/* check if address and length fit */
	if (address == 0xffffffff) {
		debug("RAM image request");
		if (length > 28l * 1024 * 1024) {
			/* image too big, reject it */
			commdebugf("RAM image too big (%d)", length);
			return FALSE;
		}
		/* the image is moved to the end of the ram, aligned on a MiB boundary */
		flashAddress = (SDRAM_END - length) & 0xfff00000;
		flashLength = length;
		flashPos = 0;
		imageRamExec = TRUE;
		imageRamAddress = (void*)flashAddress;
		imageRamLength = length;
		commdebugf("got ram address 0x%08x and length %d", flashAddress, length);
		return TRUE;
	} else if (address < COMM_FLASH_PROG_MIN || address + length > FLASH_SIZE) {
		/* address and length do not match constraints */
		commdebugf("address 0x%8x and length %d do not match constraints",
			address, length);
		return FALSE;
	} else {
		/* set start address and length */
		flashAddress = address;
		flashLength = length;
		flashPos = 0;
		commdebugf("got address 0x%08x and length %d", address, length);
		return TRUE;
	}
}

/** erase flash at ::flashAddress for at least
 *   ::flashLength bytes
 * \note this function is called as a bottom half, i.e.
 *  <b>outside Interrupt context</b> */
void eraseFlash() {
	uint32_t i;
	uint32_t progressLed, oldProgressLed = 0;
	bool justErase = FALSE;
	
	if (imageRamExec) {
		commdebugf("RAM Image, nothing to erase");
		switchState(STATE_UPLOADING);
		communicationSendPacket(&STAT_OK);
		return;
	}
	
	if (flashLength == 0) {
		justErase = TRUE;
		flashLength = FLASH_BLOCK_SIZE;
	}
	
	/* unlock flash cells and erase them */
	for (i = 0; i < flashLength; i += FLASH_BLOCK_SIZE) {
		commdebugf("erasing cell 0x%08x", flashAddress + i);
		
		/* pmic led has 8 led intensities, calculate intensity from
		   current position in flash process */
		progressLed = ((i * 8) / flashLength);
		if (i == 0 || progressLed != oldProgressLed) {
			/* first packet or progress has changed */
			oldProgressLed = progressLed;
			writePMIC(PMIC_LED1_CONTROL, (1 << 7) | (progressLed & 0x7));
		}
		flashBlockUnlock(flashAddress + i);
		if (flashBlockQueryLock(flashAddress + i)) {
			commdebugf("could not unlock cell");
			err(&STAT_ERR_UNLOCK);
			return;
		}
		if (!flashBlockErase(flashAddress + i)) {
			commdebugf("could not erase cell");
			err(&STAT_ERR_ERASE);
			return;
		}
	}
	/* turn LED off */
	writePMIC(PMIC_LED1_CONTROL, 0);
	/* switch to uploading state */
	if (justErase) {
		flashLength = 0;
	}
	switchState(STATE_UPLOADING);
	/* erase done successfully */
	communicationSendPacket(&STAT_OK);
}

/** program flash at ::flashAddress with data
 * ::flashData for ::curLength bytes
 * \note this function is called as a bottom half, i.e.
 *  <b>outside Interrupt context</b> */
void programFlash() {
	uint32_t len;
	uint32_t progressLed;
	static uint32_t oldProgressLed;
	
	if (flashLength < flashPos + curLength) {
		commdebugf("remaining length of data to be flashed (%d) < "
			"curLength (%d) at 0x%08x",
			flashLength - flashPos, curLength, flashAddress + flashPos);
		err(&STAT_ERR_PROGRAM);
	} else {
		/* pmic led has 8 led intensities, calculate intensity from
		   current position in flash process */
		progressLed = ((flashPos * 8) / flashLength);
		if (flashPos == 0 || progressLed != oldProgressLed) {
			/* first packet or progress has changed */
			oldProgressLed = progressLed;
			writePMIC(PMIC_LED1_CONTROL, (1 << 7) | (progressLed & 0x7));
		}
		
		if (imageRamExec) {
			memcpy((char*)flashAddress + flashPos, curData, curLength);
			flashPos += curLength;
		} else {
			for (; curLength != 0; curLength -= len) {
				/* program at most 64 bytes */
				len = curLength > 64 ? 64 : curLength;
				
				/* flash data, round up len to be even */
				if (!flashBlockWriteBuffered(flashAddress + flashPos,
					(uint16_t*)curData, (len / 2) + (len % 2))) {
					commdebugf("could not program");
					/* turn LED off */
					writePMIC(PMIC_LED1_CONTROL, 0);
					err(&STAT_ERR_PROGRAM);
					return;
				} else {
					/* increase data pointer */
					curData = (void*)((uint8_t*)curData + len);
					/* increase destination position */
					flashPos += len;
				}
			}
		}
		
		if (flashPos == flashLength) {
			/* last packet, turn LED off */
			writePMIC(PMIC_LED1_CONTROL, 0);
		}
		
		switchState(STATE_UPLOADING);
		communicationSendPacket(&STAT_OK);
	}
}

/** @} */

