/* Communication: Buffering
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdio.h>
#include <string.h>
#include "communication_buffer.h"
#include "communication_buffer_internal.h"
#include "lzo/minilzo.h"
#include "../debug.h"
#include "../dmalloc.h"
#include "../led.h"
#include "../config.h"
#include "../utils.h"
#include "../assert.h"

/** \defgroup communicationbuf Communication Buffer Subsystem
 * \ingroup communication
 * @{
 */
 
/** \file communication_buffer.c
 * \brief communication subsystem: configuration and buffering
 */

/** current communication system */
static struct CommunicationSystem * curSys = NULL;

static inline void processPacket(const struct PacketHeader * pkt);
static uint32_t decompress(const struct PacketHeader * pkt, void * buf,
	uint32_t len);

/** \brief Current Buffer */
static struct Buffer buf = {
	.length = 0,
	.pos = 0
};

/** \brief start communication
 * \param sys communication system to be used */
void communicationStart(struct CommunicationSystem * sys) {
	curSys = sys;
	if (curSys->start)
		curSys->start(curSys);
}

/** \brief Callback for reset
 * \note can be called from <b>Interrupt context</b> */
void communicationReset() {
#ifdef DEBUG_COMM_BUF
	debug("Communication received Reset, resetting buffer");
#endif
	ledOff(LED_BLUE);
	utilFlashLedStart();
	/* reset buffer */
	buf.pos = buf.length = 0;
}

/** \brief Callback for configured system (e.g. Bus Configuration)
 * \note can be called from <b>Interrupt context</b> */
void communicationConfigured() {
#ifdef DEBUG_COMM_BUF
	debug("Communication received Configure, resetting buffer");
#endif
	utilFlashLedStop();
	ledOn(LED_BLUE);
	/* reset buffer */
	buf.pos = buf.length = 0;
}

/** \brief Send packet
 * \note pkt must be static data as it is \b not freed afterwards */
void communicationSendPacket(const struct PacketHeader * pkt) {
	curSys->send(curSys, pkt, pkt->length);
}

/** \brief Callback when data received
 *
 * buffer it and pass to communiation system if packet is fully received
 *
 * \param data received data
 * \param length data length
 * \note can be called from <b>Interrupt context</b> */
void communicationReceive(const void * data, uint32_t length) {
	const struct PacketHeader * pkt = (const struct PacketHeader*)data;
	
	if (buf.pos == 0 && length == pkt->length) {
#ifdef DEBUG_COMM_BUF
		debugf("full packet received");
#endif
		processPacket(pkt);
	} else {
		/* append data */
		memcpy(buf.data + buf.pos, data, length);
		
		if (buf.pos == 0) {
			/* new packet received */
			buf.length = buf.pkt.length;
#ifdef DEBUG_COMM_BUF
			debugf("new packet, size: %d, now: %d, cur: %d", buf.length,
				buf.pos, length);
#endif
		} else {
			debugf("not a new packet, appended %d bytes, now: %d of %d",
				length, buf.pos + length, buf.length);
		}
		buf.pos += length;
		
		if (buf.pos == buf.length) {
			/* packet finished, process it */
#ifdef DEBUG_COMM_BUF
			debugf("packet ready, size: %d", buf.length);
#endif
			processPacket(&buf.pkt);
			/* reset buffer */
			buf.pos = 0;
			buf.length = 0;
		}
	}
}

static inline void processPacket(const struct PacketHeader * pkt) {
	if (pkt->flagCompressed) {
#ifdef DEBUG_COMM_BUF
		debugf("Packet is compressed, decompressing");
#endif
		static uint8_t data[COMM_MAX_READ_SIZE];
		uint32_t len = decompress(pkt, data, sizeof(data));
		communicationProcessPacket(pkt, data, len);
	} else {
		communicationProcessPacket(pkt, pkt->data,
			pkt->length - COMM_PKT_HEADER_SIZE);
	}
}

static uint32_t decompress(const struct PacketHeader * pkt, void * buf,
	uint32_t len) {
	lzo_uint retLen = len;
	
	/* decompress payload */
	int r = lzo1x_decompress(pkt->data,
		pkt->length - COMM_PKT_HEADER_SIZE, buf, &retLen, NULL);
	
	if (r != 0) {
		debugf("Decompression failed with code %d", r);
		error();
	}
	
	return retLen;
}

/** @} */

