/* Core Hardware: MMU management
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef MMU_H
#define MMU_H

#include <inttypes.h>

/** \brief enable the Memory Management Unit (MMU)
 * \details disables the instruction cache, sets the translation table base,
 *  activates the mmu and reenables the instruction cache (if it was enabled
 *  before)
 * \param tbl translation table base
 * \note set the domain access control register first using mmuSetDomain()
 * \note look at the Intel XScale Core Manual on Page 3.4.1-42 (MMU) and
 *  7.2.2-83 (Control Register) */
void mmuEnable(const void * tbl);

/** \brief determine wether the MMU is enabled
 * \return \c TRUE iff the MMU is enabled
 * \note look at the Intel XScale Core Manual on Page 7.2.2-83 (Control
 *  Register) */
bool mmuIsEnabled();

/** \brief disable the Memory Management Unit (MMU)
 * \details disables the instruction cache, branch target buffer and data
 *  cache, deactivates the mmu and reenables the instruction cache (if it was
 *  enabled before)
 * \note look at the Intel XScale Core Manual on Page 3.4.1-42 (MMU) and
 *  7.2.2-83 (Control Register) */
void mmuDisable();

/** \brief set the domain access control register
 * \param domain the domain to be set
 * \note look at the Intel XScale Core Manual on Page 7.2.4-85
 *  (Domain Access Control Register) */
void mmuSetDomain(uint32_t domain);

/** \brief get the domain access control register
 * \return value of the domain access control register
 * \note look at the Intel XScale Core Manual on Page 7.2.4-85
 *  (Domain Access Control Register) */
uint32_t mmuGetDomain();

/** \brief invalidate instruction tlb
 * \note this won't affect locked entries. use iTlbUnlock() if needed
 * \note look at the Intel XScale Core Manual on Page 3.4.1-42 (MMU) and
 *  7.2.9-89 (TLB Operations) */
void iTlbInvalidate();

/** \brief invalidate data tlb
 * \note this won't affect locked entries. use dTlbUnlock() if needed
 * \note look at the Intel XScale Core Manual on Page 3.4.1-42 (MMU) and
 *  7.2.9-89 (TLB Operations) */
void dTlbInvalidate();

/** \brief lock an entry in the instruction tlb
 * \param addr VMA of entry to be locked
 * \note look at the Intel XScale Core Manual on Page 3.4.3-43 (MMU) and
 *  7.2.11-91 (TLB Lock Down) */
void iTlbLockLine(const void * addr);

/** \brief lock an entry in the data tlb
 * \param addr VMA of entry to be locked
 * \note look at the Intel XScale Core Manual on Page 3.4.3-44 (MMU) and
 *  7.2.11-91 (TLB Lock Down) */
void dTlbLockLine(const void * addr);

/** \brief unlock instruction tlb
 * \note look at the Intel XScale Core Manual on Page 3.4.3-43 (MMU) and
 *  7.2.11-91 (TLB Lock Down) */
void iTlbUnlock();

/** \brief unlock data tlb
 * \note look at the Intel XScale Core Manual on Page 3.4.3-44 (MMU) and
 *  7.2.11-91 (TLB Lock Down) */
void dTlbUnlock();

#endif

