/* DMA Controller
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include <os_arm_bare_registers_def.h>
#include "dma.h"
#include "dma_internal.h"
#include "core/cache.h"
#include "dmalloc.h"
#include "debug.h"

/** \defgroup dma Direct Memory Access (DMA) Subsystem
 * @{
 */

/** \file dma.c
 * \brief Code for Direct Memory Access Controller
 *
 * Starts and Stops the DMA Controller, sets up DMA channels and serves
 *  DMA interrupts */

/** \brief Priority Sets
 * \details as defined in table 5-4 */
static const uint8_t dmaPrioritySets[4][8] = {
	[0] = {  0,  1,  2,  3, 16, 17, 18, 19 },
	[1] = {  4,  5,  6,  7, 20, 21, 22, 23 },
	[2] = {  8,  9, 10, 11, 24, 25, 26, 27 },
	[3] = { 12, 13, 14, 15, 28, 29, 30, 31 }
};

/** \brief used dma channel map.
 * \details \c TRUE iff a channel is used */
static bool dmaChannelUsed[32] = { [0 ... 31] = FALSE };

/** \brief dma channels */
static struct DmaChannel dmaChannels[32];

static void start(struct Driver * driver);
static void stop(struct Driver * driver);

/** \brief dma driver structure */
struct Driver dmaDriver = {
	.name = "DMA",
	
	.start = &start,
	.stop = &stop,
};

/** \brief start DMA */
static void start(struct Driver * driver) {
	/* enable dma interrupt */
	ICMR_REG(IID_DMAC) |= INTERRUPT_BIT(IID_DMAC);
}

/** \brief stop DMA */
static void stop(struct Driver * driver) {
	uint8_t channel;
	
	/* disable dma interrupt */
	ICMR_REG(IID_DMAC) &= ~INTERRUPT_BIT(IID_DMAC);
	
	/* ack pending interrupts */
	for (channel = 0; channel < 32; ++channel)
		DCSR(channel) |= 0;
}

/** \brief DMA Interrupt Handler
 *
 * called by main interrupt handler whenever a dma interrupt occures */
void dmaInterrupt() {
	uint32_t dint;
	uint8_t channel;
	
	/* read dma interrupt register */
	dint = DINT;
	channel = 0;
	while (dint) {
		if ((dint & 0x1) && dmaChannelUsed[channel]) {
			/* dma interrupt on channel */
			uint32_t dcsr = DCSR(channel);
			/* stop dma channel and clear interrupt */
			DCSR(channel) = (dcsr & ~(DCSR_RUN | DCSR_STOPIRQEN))
				| DCSR_ENDINTR | DCSR_EORINT;
#ifdef DEBUG_DMA
			debugf("dma INTR on ch %d, 0x%08x 0x%08x, L: %d", channel, dcsr, DCSR(channel),
				DCMD(channel) & 0x1fff);
#endif
			/* deliver data */
			dmaChannels[channel].cbDmaInterrupt(channel,
				DCMD(channel) & 0x1fff, dmaChannels[channel].data);
		}
		/* next channel and interrupt */
		channel++;
		dint >>= 1;
	}
}

/** \brief allocate new channel
 *
 * picks up an unused DMA channel with the given priority and marks it as
 *  used.
 *
 * \param priority priority of the requested dma channel
 * \param[out] channel the allocated channel, valid iff return value is true
 * \param drcm dma request mapping
 * \param cbDmaInterrupt callback on dma interrupt
 * \param data user defined data to be passed to the callback
 *
 * \return \c TRUE iff the channel could be allocated */
bool dmaGetChannel(uint8_t priority, uint8_t * channel, uint32_t drcm,
	void (*cbDmaInterrupt)(uint8_t channel, uint32_t remainingLength,
		void * data), void * data) {
	const uint8_t * prioritySet;
	uint8_t i, ch;
	
	if (priority > 3 || channel == NULL) return FALSE;
	/* get the requested priority set */
	prioritySet = dmaPrioritySets[priority];
	/* iterate over set */
	for (i = 0; i < 8; ++i) {
		/* search for next free channel */
		ch = prioritySet[i];
		if (!dmaChannelUsed[ch]) {
			/* found an unused channel - setup and return true */
			dmaChannelUsed[ch] = TRUE;
			/* map request to channel */
			DRCMR(drcm) = DRCMR_CHLNUM(ch) | DRCMR_MAPVLD;
			dmaChannels[ch].drcm = drcm;
			dmaChannels[ch].cbDmaInterrupt = cbDmaInterrupt;
			dmaChannels[ch].data = data;
			*channel = ch;
			return TRUE;
		}
	}
	/* no such channel found, allocation failed */
	return FALSE;
}

/** \brief release a previously allocated channel
 * \param channel DMA channel as returned by dmaGetChannel()
 * \see dmaGetChannel */
void dmaReleaseChannel(uint8_t channel) {
	dmaKillChannel(channel);
	/* mark the channel as unused */
	dmaChannelUsed[channel] = FALSE;
}

/** \brief stop DMA channel
 * \param channel DMA channel */
void dmaKillChannel(uint8_t channel) {
	/* stop channel */
	DCSR(channel) &= ~(DCSR_RUN | DCSR_EORIRQEN | DCSR_STOPIRQEN);
	/* unmap request */
	DRCMR(dmaChannels[channel].drcm) &= ~DRCMR_MAPVLD;
}

/** \brief stop DMA channel
 * \param channel DMA channel */
void dmaStopChannel(uint8_t channel) {
	/* stop channel */
	DCSR(channel) &= ~DCSR_RUN;
}

void dmaRemapChannel(uint8_t channel) {
	/* remap request */
	DRCMR(dmaChannels[channel].drcm) |= DRCMR_MAPVLD;
}

/** \brief setup a DMA channel for transferring data
 *  from a hardware fifo to memory
 *
 * \param channel DMA channel as returned by dmaGetChannel()
 * \param sourceReg the fifo register to be transferred from
 * \param destMem the memory location to be transferred to
 * \param size burst size of the dma transfer
 * \param width width of the dma transfer
 * \param len length of the destination memory */
void dmaSetupHWToMem(uint8_t channel, const void * sourceReg, void * destMem,
	uint8_t size, uint8_t width, uint32_t len) {
	
	/* no descriptor fetch */
	DCSR(channel) = DCSR_NODESCFETCH;
	/* source address */
	DSADR(channel) = (uint32_t)sourceReg;
	/* target address */
	DTADR(channel) = (uint32_t)destMem;
	/* increment target address, source flow, irq on end of flow,
		size, width and length */
	DCMD(channel) = DCMD_INCTRGADDR | DCMD_FLOWSRC | DCMD_ENDIRQEN |
		DCMD_SIZE(size) | DCMD_WIDTH(width) | DCMD_LEN(len);
	/* irq on end of request, start dma channel */
	/* TODO: this is what usb needs, maybe we should parameterize that */
	/* | DCSR_EORSTOPEN;DCSR_EORIRQEN;*/
	DCSR(channel) = DCSR_NODESCFETCH | DCSR_RUN | DCSR_STOPIRQEN;
}

/** \brief setup a DMA channel for transferring data
 *  from memory to a hardware fifo
 *
 * \param channel DMA channel as returned by dmaGetChannel()
 * \param sourceMem the memory location to be transferred from
 * \param destReg the fifo register to be transferred to
 * \param size burst size of the dma transfer
 * \param width width of the dma transfer
 * \param len length of the destination memory */
void dmaSetupMemToHW(uint8_t channel, const void * sourceMem, void * destReg,
	uint8_t size, uint8_t width, uint32_t len) {
	
	/* no descriptor fetch */
	DCSR(channel) = DCSR_NODESCFETCH;
	/* source address */
	DSADR(channel) = (uint32_t)sourceMem;
	/* target address */
	DTADR(channel) = (uint32_t)destReg;
	/* increment source address, target flow, irq on end of flow,
		size, width and length */
	DCMD(channel) = DCMD_INCSRCADDR | DCMD_FLOWTRG | DCMD_ENDIRQEN |
		DCMD_SIZE(size) | DCMD_WIDTH(width) | DCMD_LEN(len);
	
	/* irq on end of request, start dma channel */
	DCSR(channel) = DCSR_NODESCFETCH | DCSR_RUN | DCSR_EORIRQEN;
}

struct DmaDescriptor * dmaAllocateDescriptor() {
	/* allocate descriptor to a 32 byte alignment */
	return (struct DmaDescriptor*)dmemalign(32, sizeof(struct DmaDescriptor));
}

/** \brief setup a DMA descriptor for transferring data
 *  from memory to a hardware fifo
 *
 * \param [out] desc dma descriptor to be set up, must be aligned correctly
 * \param stop \c TRUE iff this is the last descriptor in a descriptor chain
 * \param sourceMem the memory location to be transferred from
 * \param destReg the fifo register to be transferred to
 * \param size burst size of the dma transfer
 * \param width width of the dma transfer
 * \param len length of the destination memory
 * \note the returned descriptor must be freed after the request is done.
 *  This should be done by using dmaFreeDescriptorChain()
 * \see dmaFreeDescriptorChain
 * \see dmaSetupDescStart
 */
void dmaSetupDescMemToHW(struct DmaDescriptor * desc, bool stop,
	const void * sourceMem, void * destReg, uint8_t size, uint8_t width,
	uint32_t len) {
	/* setup descriptor */
	desc->ddadr = (void*)(stop ? DDADR_STOP : 0);
	desc->dsadr = sourceMem;
	desc->dtadr = destReg;
	desc->dcmd = DCMD_INCSRCADDR | DCMD_FLOWTRG | DCMD_ENDIRQEN |
		DCMD_SIZE(size) | DCMD_WIDTH(width) | DCMD_LEN(len);
}

/** \brief setup a DMA descriptor for transferring data
 *  from memory to a hardware fifo
 *
 * \param [out] desc dma descriptor to be set up, must be aligned correctly
 * \param stop \c TRUE iff this is the last descriptor in a descriptor chain
 * \param sourceReg the fifo register to be transferred from
 * \param destMeM the memory location to be transferred to
 * \param size burst size of the dma transfer
 * \param width width of the dma transfer
 * \param len length of the destination memory
 * \note the returned descriptor must be freed after the request is done.
 *  This should be done by using dmaFreeDescriptorChain()
 * \see dmaFreeDescriptorChain
 * \see dmaSetupDescStart
 */
void dmaSetupDescHWToMem(struct DmaDescriptor * desc, bool stop,
	const void * sourceReg, void * destMem, uint8_t size, uint8_t width,
	uint32_t len) {
	/* setup descriptor */
	desc->ddadr = (void*)(stop ? DDADR_STOP : 0);
	desc->dsadr = sourceReg;
	desc->dtadr = destMem;
	desc->dcmd = DCMD_INCTRGADDR | DCMD_FLOWSRC | DCMD_ENDIRQEN |
		DCMD_SIZE(size) | DCMD_WIDTH(width) | DCMD_LEN(len);
}

/** \brief setup a DMA descriptor for transferring data from memory to memory
 *
 * \param [out] desc dma descriptor to be set up, must be aligned correctly
 * \param stop \c TRUE iff this is the last descriptor in a descriptor chain
 * \param sourceMem the memory location to be transferred from
 * \param destMem the memory location to be transferred to
 * \param size burst size of the dma transfer
 * \param len length of the destination memory
 * \return address of the new descriptor or \c NULL if allocation failed
 * \note the returned descriptor must be freed after the request is done.
 *  This should be done by using dmaFreeDescriptorChain()
 * \see dmaFreeDescriptorChain
 * \see dmaSetupDescStart */
void dmaSetupDescMemToMem(struct DmaDescriptor * desc,
	bool stop, const void * sourceMem, void * destMem, uint8_t size,
	uint32_t len) {
	/* setup descriptor */
	desc->ddadr = (void*)(stop ? DDADR_STOP : 0);
	desc->dsadr = sourceMem;
	desc->dtadr = destMem;
	desc->dcmd = DCMD_INCSRCADDR | DCMD_INCTRGADDR | DCMD_ENDIRQEN |
		DCMD_SIZE(size) | DCMD_WIDTH(0) | DCMD_LEN(len);
}

/** \brief set next DMA descriptor in a descriptor chain
 * \param desc descriptor
 * \param next next descriptor
 */
void dmaSetupDescSetNextDesc(struct DmaDescriptor * desc,
	const struct DmaDescriptor * next) {
	desc->ddadr = (struct DmaDescriptor*)
		((uint32_t)DDADR_ADDR(next) | DDADR_FLAGS(desc->ddadr));
}

/** \brief start DMA transfer from a DMA descriptor
 *
 * \param channel DMA channel as returned by dmaGetChannel()
 * \param desc DMA descriptor as setup by dmaSetupDescMemToHW() or
 *  dmaSetupDescMemToMem()
 * \see dmaSetupDescMemToHW
 * \see dmaSetupDescMemToMem */
void dmaSetupDescStart(uint8_t channel, const struct DmaDescriptor * desc) {
	struct DmaDescriptor * next, * addr;
	uint32_t flags;
	
	/* invalidate caches of all descriptors */
	next = (struct DmaDescriptor*)desc;
	do {
		addr = DDADR_ADDR(next);
		
		dCacheCleanAndInvalidateLine(addr, sizeof(*addr));
		
		flags = DDADR_FLAGS(addr->ddadr);
		next = addr->ddadr;
	} while(!(flags & DDADR_STOP) && DDADR_ADDR(next) != NULL);
	
	/* setup channel */
	DCSR(channel) &= ~(DCSR_RUN | DCSR_NODESCFETCH);
	DDADR(channel) = (uint32_t)DDADR_ADDR(desc);
	/* start dma channel */
	DCSR(channel) = DCSR_RUN | DCSR_EORIRQEN;
}

/** \brief free DMA descriptor chain
 *
 * \param desc first descriptor of chain
 * \see dmaSetupDescMemToHW
 * \see dmaSetupDescMemToMem */
void dmaFreeDescriptorChain(struct DmaDescriptor * desc) {
	struct DmaDescriptor * addr, * next;
	uint32_t flags;
	
	/* free all descriptors */
	next = desc;
	do {
		addr = DDADR_ADDR(next);
		
		flags = DDADR_FLAGS(addr->ddadr);
		next = addr->ddadr;
		dfree(addr);
	} while(!(flags & DDADR_STOP) && DDADR_ADDR(next) != NULL);
}

/** @} */

