/* USB Driver: Interrupt handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <os_arm_bare_registers_def.h>
#include <inttypes.h>
#include "usb_control_pipe_internal.h"
#include "usb_main_internal.h"
#include "usb_configuration_internal.h"
#include "usb_endpoint_internal.h"
#include "usb_debug.h"

/** \defgroup usbintr USB Interrupt Handler
 * \ingroup usb
 * @{
 */

/** \file usb_interrupt.c
 * \brief USB Interrupt Handler
 */

static void busResetInterrupt();
static void busReconfigInterrupt();

/** \brief called whenever a UDC Interrupt is fired */
void usbInterrupt() {
	uint32_t udcisr0 = UDCISR0;
	uint32_t udcisr1 = UDCISR1;
	
	usbDebug("Interrupt, UDCISR0:0x%08x, UDCISR1:0x%08x,UDCCR:0x%08x",
		udcisr0, udcisr1, UDCCR);
	
	if (udcisr1 & UDCISR_IRSU) {
		/* suspend */
	}
	if (udcisr1 & UDCISR_IRRU) {
		/* resume */
	}
	if (udcisr1 & UDCISR_IRCC) {
		busReconfigInterrupt();
	}
	if (udcisr1 & UDCISR_IRRS) {
		busResetInterrupt();
	}
	
	if ((udcisr0 & UDCISR0_EP_MASK) | (udcisr1 & UDCISR1_EP_MASK)) {
		usbEndpointInterrupt();
	}
}

/** \cond INTERNAL */

/** \brief called on a bus reset
 *
 * stops all requests, unconfigures udc interrupts and calls the callback */
static void busResetInterrupt() {
	uint32_t udccr = UDCCR;
	
	/* ack irq */
	UDCISR1 = UDCISR_IRRS;
	
	if ((udccr & UDCCR_UDA) == 0) {
		/* there is no activity on the bus */
		usbInfo("Bus Reset: No activity on bus");
		usbEndpointsFinishAll(USB_REQUEST_STATUS_RESET);
	} else {
		usbInfo("Bus Reset: Activity on bus");
	}
	
	/* reset endpoint 0 */
	usbEndpointNuke(&usbEndpoint0, USB_REQUEST_STATUS_RESET);
	usbWrite_UDCCSR(&usbEndpoint0, UDCCSR0_FTF | UDCCSR0_OPC);
	usbControlPipeSetState(USB_CP_WAIT_FOR_SETUP);
	
	/* call callback */
	if (usbUdc.config->cbBusReset)
		usbUdc.config->cbBusReset(usbUdc.config);
	
	usbUdc.configIdx = 0;
	usbUdc.ifaceIdx = 0;
	usbUdc.settingIdx = 0;
}

/** \brief called on a SET_CONFIGURATION request
 *
 * Enables the requested configuration and checks if it has really changed.
 *  If so, the UDC interrupts are reconfigured and the configuration callback
 *  is called. */
void busReconfigInterrupt() {
	uint32_t udccr;
	uint8_t configIdx, ifaceIdx, settingIdx;
	
	udccr = UDCCR;
	
	/* ack interrupt */
	UDCISR1 = UDCISR_IRCC;
	
	/* read new configuration, interface and setting index */
	configIdx = UDCCR_ACN(udccr);
	ifaceIdx = UDCCR_AIN(udccr);
	settingIdx = UDCCR_AAISN(udccr);
	
	usbInfo("Bus Configure: Config %d, Iface %d, Setting %d",
		configIdx, ifaceIdx, settingIdx);
	
	if (configIdx == 0 || configIdx > usbUdc.config->numConfigurations) {
		usbErr("Configuration Index %d out of range (1..%d)\n", configIdx,
			usbUdc.config->numConfigurations);
		return;
	} else if(ifaceIdx >=
		usbUdc.config->configurations[configIdx - 1]->numInterfaces) {
		usbErr("Interface Index %d out of range (0..%d)\n", ifaceIdx,
			usbUdc.config->configurations[configIdx - 1]->numInterfaces - 1);
		return;
	} else if(settingIdx >
		usbUdc.config->configurations[configIdx - 1]->interfaces[ifaceIdx]->numAltSettings) {
		usbErr("Altnernate Setting Index %d out of range (0..%d)\n",
			settingIdx,
			usbUdc.config->configurations[configIdx - 1]->interfaces[ifaceIdx]->numAltSettings - 1);
		return;
	}
	
	if (configIdx != usbUdc.configIdx || ifaceIdx != usbUdc.ifaceIdx ||
		settingIdx != usbUdc.settingIdx) {
		/* change in configuration */
		
		struct UsbAltSetting * newAltSetting =
			usbUdc.config->configurations[configIdx - 1]->interfaces[ifaceIdx]->altSettings[settingIdx];
		
		/* configure endpoints with new configuration (also finishes the old
		 * endpoint configuration) */
		usbEndpointsConfigure(newAltSetting);
		
		usbUdc.curAltSetting = newAltSetting;
		
		usbUdc.configIdx = configIdx;
		usbUdc.ifaceIdx = ifaceIdx;
		usbUdc.settingIdx = settingIdx;
		
		/* call callback */
		if (usbUdc.config->cbConfigured) {
			usbUdc.config->cbConfigured(usbUdc.config, configIdx - 1, ifaceIdx,
				settingIdx);
		}
	}
	
	/* enable memory configuration */
	usbSet_UDCCR(UDCCR_SMAC);
}

/** \endcond */

/** @} */

